@@ -3,7 +3,7 @@
 #
 #  (C) Paul Evans, 2015-2021 -- leonerd@leonerd.org.uk
 
-package Device::Chip::Adapter::BusPirate 0.21;
+package Device::Chip::Adapter::BusPirate 0.22;
 
 use v5.14;
 use warnings;
@@ -12,6 +12,7 @@ use base qw( Device::Chip::Adapter );
 use Carp;
 
 use Future::AsyncAwait;
+use Future::Mutex;
 
 use Device::BusPirate;
 
@@ -127,7 +128,9 @@ async sub make_protocol_I2C
 
    my $mode = await $self->_enter_mode_I2C;
 
-   return Device::Chip::Adapter::BusPirate::_I2C->new( $mode );
+   my $mutex = $self->{mutex} //= Future::Mutex->new;
+
+   return Device::Chip::Adapter::BusPirate::_I2C->new( $mode, $mutex );
 }
 
 async sub make_protocol_UART
@@ -161,9 +164,9 @@ use List::Util qw( first );
 sub new
 {
    my $class = shift;
-   my ( $mode ) = @_;
+   my ( $mode, $mutex ) = @_;
 
-   bless { mode => $mode }, $class;
+   bless { mode => $mode, mutex => $mutex }, $class;
 }
 
 sub sleep
@@ -383,13 +386,7 @@ async sub write
    my $self = shift;
    my ( $bytes ) = @_;
 
-   defined( my $addr = $self->{addr} ) or
-      croak "Cannot ->write without a defined addr";
-
-   return await $self->txn(sub { $self->write( $bytes ) }) unless $self->{in_txn};
-
-   await $self->{mode}->start_bit;
-   await $self->{mode}->write( chr( $addr << 1 | 0 ) . $bytes );
+   await $self->txn(sub { shift->write( $bytes ) });
 }
 
 async sub read
@@ -397,31 +394,7 @@ async sub read
    my $self = shift;
    my ( $len ) = @_;
 
-   defined( my $addr = $self->{addr} ) or
-      croak "Cannot ->read without a defined addr";
-
-   return await $self->txn(sub { $self->read( $len ) }) unless $self->{in_txn};
-
-   await $self->{mode}->start_bit;
-   await $self->{mode}->write( chr( $addr << 1 | 1 ) );
-   return await $self->{mode}->read( $len );
-}
-
-sub txn
-{
-   my $self = shift;
-   my ( $code ) = @_;
-
-   my $mutex = $self->{txn_mutex} //= Future::Mutex->new;
-
-   return $mutex->enter(sub {
-      $self->{in_txn} = 1;
-      return $code->()->followed_by(sub {
-         my ( $f ) = @_;
-         $self->{in_txn} = 0;
-         return $self->{mode}->stop_bit->then( sub { $f } );
-      });
-   });
+   return await $self->txn(sub { shift->read( $len ) });
 }
 
 async sub write_then_read
@@ -430,11 +403,54 @@ async sub write_then_read
    my ( $write_bytes, $read_len ) = @_;
 
    return await $self->txn(async sub {
-      await $self->write( $write_bytes );
-      return await $self->read( $read_len );
+      my ( $helper ) = @_;
+      await $helper->write( $write_bytes );
+      return await $helper->read( $read_len );
    });
 }
 
+sub txn
+{
+   my $self = shift;
+   my ( $code ) = @_;
+
+   defined( my $addr = $self->{addr} ) or
+      croak "Cannot ->txn without a defined addr";
+
+   my $helper = $self->{txn_helper} //= bless( [ $self->{mode}, $self->{addr} ], "Device::Chip::Adapter::BusPirate::_I2C::Txn" );
+
+   return $self->{mutex}->enter(sub {
+      return $code->( $helper )->followed_by(sub {
+         my ( $f ) = @_;
+         return $self->{mode}->stop_bit->then( sub { $f } );
+      });
+   });
+}
+
+package
+   Device::Chip::Adapter::BusPirate::_I2C::Txn;
+
+async sub write
+{
+   my $self = shift;
+   my ( $bytes ) = @_;
+   my ( $mode, $addr ) = @$self;
+
+   await $mode->start_bit;
+   await $mode->write( chr( $addr << 1 | 0 ) . $bytes );
+}
+
+async sub read
+{
+   my $self = shift;
+   my ( $len ) = @_;
+   my ( $mode, $addr ) = @$self;
+
+   await $mode->start_bit;
+   await $mode->write( chr( $addr << 1 | 1 ) );
+   return await $mode->read( $len );
+}
+
 package
    Device::Chip::Adapter::BusPirate::_UART;
 use base qw( Device::Chip::Adapter::BusPirate::_base );
